#include <stdio.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include "mat.h"
#include "odometer.h"    /* All external variables */
#include "odometer_p.h"

#include "lasutil.h"
#include "err_integration_p.h"

/* #include "odometer_sick.h" */
 
/***************
this file deals with robot'encoders and
the position deduced from the sensors.

we call : rob_enc[4] the last known robot position given
          by the encoders

          Glob_pos[4] the last update of the robot position 
          in the absolute coordinates system
          taking into account the sensor data

          Loc2Abs[3] the coordinates of the starting position of
          the robot (x,y,th) in the absolute frame chosen by the user.

          Loc2Abs[3] : projection of rob_enc = (0,0,0) in the global coordinate
           system.

 General rules: rob_enc is never to be updated by the user. It is updated by the robot.
                Generally, the user   updates Glob_pos, typically after a relocalization. 
		   The system computes then Loc2Abs such that rob_enc 
		   remains consistent. The function to use are then 
                                                   init_location_st or init_location_tu.
		   
    There are still a number of issues around here . Especially if we want to use the variance 
    provided on the robot side. How do we project it on the server side ? 

    I we set Vrob_enc directly as showed in get_location_err(), the resulting variance will be way to big 
    because it doesn't take into account the correlation between Vrob_enc and Vloc2abs.  I see only two 
    ways here : 

    - the robot not only provides the variance matrix of the current Rob_enc position, but also the covariance
    matrix with respect to the previous time Vloc2abs has been computed

    - the robot variance is set to NULL when a Vloac2abs is performed and then the variance of the current
    robot position will be recomputed from there.

    The first solution is probably cleaner as it doesn't interfere with the odometric system on the robot. 

    Updated : the first solution has been implemented using a little trick described in 
    odometer.tex



    But warning : set_encoders should not be called before a init_encoders has been performed !
****************/

/** The following public variables are declared in "odometer.h" **/

/* RealVariance :

   RealVariance is to be set to 1 if the robot has a probabilistic odometry,
   or to 0 otherwise. In this second case the variances are estimated by segkit
   (for diligent for ionstance).
   This value has to be initialized in function InitSick (file sick_<robot>.h).
   (here it is only a default value).
*/
int RealVariance = 0;

/****** Public:

  Robot position and variance in absolute coord. sys. 
  To acces these variable, include the file "rob_pos.h"
******/
 
double Glob_pos[4] = {0.,0.,0.,0.}; /* position */
double Vpos[6] = {0.,0.,0.,0.,0.,0.}; /* Variance on the position */

/******* public:
  starting position (and its variance) of the
  robot in the choosen absolute frame; By default,
  the absolute frame is located at the starting position of the robot
********/
  
double Loc2Abs[3] = {0.,0.,0.};
double Vloc2abs[6] = {0.,0.,0.,0.,0.,0.};



/******* public: *********/

double rob_enc[4] = {0.,0.,0.,0.}; /* position given by encoders */
double Vrob_enc[6] = {0.,0.,0.,0.,0.,0.}; /* variance on the encoders integration */

#define deg2rad(x) (3.141592653589793 * (x) /180.)
#define rad2deg(x) ((x)*180/3.141592653589793)


static double cloc2abs = 1.; /* cosine of Loc2Abs[2] */
static double sloc2abs = 0.; /* sine of Loc2Abs[2] */




double Prev_pos[4];
double prev_odo[3];

static int update_encoders (double *pos);

/*** set the encoder values from pos
     recomputes Glob_pos from Loc2Abs.

 **/

static char InitEncoderOK = 0;

void set_encoders(double *pos, double *var)
{

  

  update_encoders(pos);

  cloc2abs = cos(Loc2Abs[2]);
  sloc2abs = sin(Loc2Abs[2]);


  Glob_pos[0] = cloc2abs * rob_enc[0] - sloc2abs * rob_enc[1] + Loc2Abs[0]; 
  Glob_pos[1] = sloc2abs * rob_enc[0] + cloc2abs * rob_enc[1] + Loc2Abs[1];
  Glob_pos[2] = norm_ang(rob_enc[2] + Loc2Abs[2],-M_PI) ;
  Glob_pos[3] = norm_ang(rob_enc[3] + Loc2Abs[2],-M_PI) ;


  /**** 
        sim: computing Vrob_enc;
        in real replace by get_location_err()
  *****/

  if (RealVariance)
  
      /*NOTE : in this case, one HAS to provide this function in a separate
	file like it is done in each file in the directory ../../robots
      */
  {    
    int i;
    for(i=0;i<6;i++)
      Vrob_enc[i] = var[i];
  }
  
  

  /* projection dans le repere de global loc */
  get_location_err();

}

/* separate function for just updating encoder values whenever 
   we get it. In case of simulated variance, the new variance is computed also.
   This allows to integrate errors even we don't proceed to a localization */

static int update_encoders (double *pos)
{
    double trans_com, rot_com;
#if VERBOSE>1
    double toto[6];
#endif

  if( !InitEncoderOK){
    printf("Error in set_encoders : init_encoders didn't occur !\n");
    return 0;
  }
  

  rob_enc[0] = pos[0];
  rob_enc[1] = pos[1];
  rob_enc[2] = pos[2];
  rob_enc[3] = pos[3];
    
  /**** 
        sim: computing Vrob_enc;
        in real replace by get_location_err()
  *****/

  if (!RealVariance){

  trans_com = sqrt((prev_odo[0] - rob_enc[0])*(prev_odo[0] - rob_enc[0]) +
		   (prev_odo[1] - rob_enc[1])*(prev_odo[1] - rob_enc[1]));

  rot_com = norm_ang(rob_enc[2] - prev_odo[2],-M_PI);

  integrate_error(trans_com, rot_com);
  
  }
  mx_copy (rob_enc,prev_odo,3,1);
}


/** 
  sets the current value of Glob_pos and Vpos to zero
  reinitialize  the encoders to the value xp;
  sets Vrob_enc to 0;
  computes the transform Loc2Abs and Vloc2abs

  Mainly useful for starting a new map.
**/
  
void init_encoders(double *xp)
{
  
  InitEncoderOK = 1;

  /* this is weird, I know ... 
     Just linking variables from this file to the ones unsed in err_integration.c */

  LinkExternalVariables();

  rob_enc[0] = xp[0];
  rob_enc[1] = xp[1];
  rob_enc[2] = xp[2];
  rob_enc[3] = xp[3];

  prev_odo[0] = xp[0];
  prev_odo[1] = xp[1];
  prev_odo[2] = xp[2];

  printf("Set prev_odo to %.2f %.2f\n",xp[0],xp[1]);

  Glob_pos[0] = 0.; 
  Glob_pos[1] = 0.;
  Glob_pos[2] = 0.;
  Glob_pos[3] = xp[3] - xp[2];

  /* zeroing the variance on the map side */

  mxs_init(Vpos,3,3,0.);

  /** 
      OUTPHIL
      zeroing the variance on the robot side : changed !
      now, Vloc2Abs should always reflect the current variance on the 
      robot as well. For now, we do not want to change it physically 
      on the robot. So this goes ...     
   **/
  /*mxs_init(Vloc2abs,3,3,0.);
    Init_variance();*/
  
  
  CalcTransf(Glob_pos,Vpos,rob_enc,Loc2Abs,Vloc2abs);  
  
}

void init_glob_pos(double *xp)
{
  InitEncoderOK = 1;

  /* this is weird, I know ... 
     Just linking variables from this file to the ones unsed in err_integration.c */

  LinkExternalVariables();

  rob_enc[0] = xp[0];
  rob_enc[1] = xp[1];
  rob_enc[2] = xp[2];
  rob_enc[3] = xp[3];

  prev_odo[0] = xp[0];
  prev_odo[1] = xp[1];
  prev_odo[2] = xp[2];

  printf("Set prev_odo to %.2f %.2f\n",xp[0],xp[1]);

  Glob_pos[0] = xp[0]; 
  Glob_pos[1] = xp[1];
  Glob_pos[2] = xp[2];
  Glob_pos[3] = xp[3];

  /* zeroing the variance on the map side */

  mxs_init(Vpos,3,3,0.);

  /** 
      OUTPHIL
      zeroing the variance on the robot side : changed !
      now, Vloc2Abs should always reflect the current variance on the 
      robot as well. For now, we do not want to change it physically 
      on the robot. So this goes ...     
   **/
  /*mxs_init(Vloc2abs,3,3,0.);
    Init_variance();*/
  
  
  CalcTransf(Glob_pos,Vpos,rob_enc,Loc2Abs,Vloc2abs);  

}
   
/***********
  init location from an array of 3 doubles defining the new position
  of x,y and the STEERING angle ; 
  the location for the turret is set to steering + rob_enc[3] - rob_enc[4].
  
  One should use this function when the local data used for relocation
  was expressed in the steering coordinate system

  Recomputes Loc2Abs and its variance.
  If you don't want to deal with the variance, put vxp = null
  OUTPHIL, comment removed "In any case, zeroes the variance on the robot side."

  Mainly used when no transformation has been computed (partial localization in 
  corridors for example).

  Here, it is time to compute the jacobian Jac_Loc2Abs_PreviousEncPos
***********/



void init_location_st(double *x, double *vxp)
{
  
    
  Glob_pos[0] = x[0]; 
  Glob_pos[1] = x[1];
  Glob_pos[2] = x[2];
  Glob_pos[3] = norm_ang(x[2] - rob_enc[2] + rob_enc[3],-M_PI);

  /** zeroing on the robot side
      OUTPHIL : nope anymore
   **/
  /*Init_variance();*/

  if(vxp != NULL){
    mxs_copy(vxp,Vpos,3,3);
    CalcTransf(Glob_pos,vxp,rob_enc,Loc2Abs,Vloc2abs);
  }
  else{
   CalcTransfNp(Glob_pos,rob_enc,Loc2Abs);
  }
  /* jacobian */
  Jac_Loc2Abs_PreviousEncPos();
  /* storing the variance Vrob_enc corresponding to Vloc2abs */
  SavePreviousEncoderVariance();
  Prev_pos[0] = rob_enc[0];
  Prev_pos[1] = rob_enc[1];
  Prev_pos[2] = rob_enc[2];
}



/**** take encoder values; computes Glob_pos *****/

/****
  computes g_pos according to encoder values in enc and the current 
  Loc2Abs transform
****/

void GetAbsPos(double *enc,double *g_pos)
{
   Calc_abs_pos(enc,NULL,Loc2Abs,NULL,g_pos,NULL);

} 

/*** 
     takes a transform x and its variance
     copies x in Loc2Abs 
     copies vtr in Vloc2abs     
     recomputes the global position Glob_pos
     recomputes Vpos

     Used when a full localization has occured.

     Here again, time for a Jac_Loc2Abs_PreviousEncPos
*****/

void init_tr_st(double *x,double *vtr)
{
  
  
  Loc2Abs[0] = x[0];
  Loc2Abs[1] = x[1];
  Loc2Abs[2] = x[2];
 
  
 
  if(vtr != NULL){
    mxs_copy(vtr,Vloc2abs,3,3);
    
  }
  /* CVS 1.11 :
     note that I do not recompute Vpos here. Vpos is computed only by 
     LocalizationPositionVariance() now */
  Calc_abs_pos(rob_enc,NULL,x,NULL,Glob_pos,NULL);

  /* jacobian */
  Jac_Loc2Abs_PreviousEncPos();
  /* storing the variance Vrob_enc corresponding to Vloc2abs */
  SavePreviousEncoderVariance();
  Prev_pos[0] = rob_enc[0];
  Prev_pos[1] = rob_enc[1];
  Prev_pos[2] = rob_enc[2];
}


 
/***********
  computes  the current location error into 
  the array Vpos; 
  It is used 
  after some fonction has set Vrob_enc to whatever the variance 
  is on the robot side. 
****************/

void get_location_err(void)
{
  LocalizationPositionVariance();
}













